# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm CLI
"""

import logging
import time

import typer
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from wuttafarm.cli import wuttafarm_typer


log = logging.getLogger(__name__)


class ChangeProcessor:

    def __init__(self, config):
        self.config = config
        self.app = config.get_app()

    def process_change(self, change):
        if change.deleted:
            self.delete_record(change)
        else:
            self.import_record(change)

    def import_record(self, change):
        token = self.get_farmos_oauth2_token()
        client = self.app.get_farmos_client(token=token)

        full_type = f"{change.entity_type}--{change.bundle}"
        record = client.resource.get_id(
            change.entity_type, change.bundle, change.farmos_uuid
        )

        importer_map = self.get_importer_map()
        model_name = importer_map[full_type]

        self.app.auto_sync_from_farmos(record["data"], model_name, client=client)

    def delete_record(self, change):
        model = self.app.model
        handler = self.app.get_import_handler("import.to_wuttafarm.from_farmos")

        importer_map = self.get_importer_map()
        full_type = f"{change.entity_type}--{change.bundle}"
        model_name = importer_map[full_type]

        token = self.get_farmos_oauth2_token()
        client = self.app.get_farmos_client(token=token)

        # nb. begin txn to establish the API client
        handler.begin_source_transaction(client)
        with self.app.short_session(commit=True) as session:
            handler.target_session = session
            importer = handler.get_importer(model_name, caches_target=False)

            # try to attribute change to 'farmos' user
            if user := session.query(model.User).filter_by(username="farmos").first():
                session.info["continuum_user_id"] = user.uuid

            # only support importers with farmos_uuid as key
            # (pretty sure that covers us..can revise if needed)
            if importer.get_keys() != ["farmos_uuid"]:
                log.warning(
                    "unsupported keys for %s importer: %s",
                    model_name,
                    importer.get_keys(),
                )
                return

            # delete corresponding record from our app
            obj = importer.get_target_object((change.uuid,))
            if obj:
                importer.delete_target_object(obj)

    # TODO: this should live elsewhere
    def get_farmos_oauth2_token(self):

        client_id = self.config.get(
            "farmos.oauth2.importing.client_id", default="wuttafarm"
        )
        client_secret = self.config.require("farmos.oauth2.importing.client_secret")
        scope = self.config.get("farmos.oauth2.importing.scope", default="farm_manager")

        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)

        return oauth.fetch_token(
            token_url=self.app.get_farmos_url("/oauth/token"),
            include_client_id=True,
            client_secret=client_secret,
            scope=scope,
        )

    # TODO: this should live elsewhere
    def get_importer_map(self):
        return {
            "asset--animal": "AnimalAsset",
            "asset--equipment": "EquipmentAsset",
            "asset--group": "GroupAsset",
            "asset--land": "LandAsset",
            "asset--plant": "PlantAsset",
            "asset--structure": "StructureAsset",
            "asset--water": "WaterAsset",
            "log--activity": "ActivityLog",
            "log--harvest": "HarvestLog",
            "log--medical": "MedicalLog",
            "log--observation": "ObservationLog",
            "log--seeding": "SeedingLog",
            "quantity--material": "MaterialQuantity",
            "quantity--standard": "StandardQuantity",
            "taxonomy_term--animal_type": "AnimalType",
            "taxonomy_term--equipment_type": "EquipmentType",
            "taxonomy_term--plant_type": "PlantType",
            "taxonomy_term--season": "Season",
            "taxonomy_term--material_type": "MaterialType",
            "taxonomy_term--unit": "Unit",
        }


@wuttafarm_typer.command()
def process_webhooks(
    ctx: typer.Context,
):
    """
    Process incoming webhook requests from farmOS.
    """
    config = ctx.parent.wutta_config
    app = config.get_app()
    model = app.model
    processor = ChangeProcessor(config)

    while True:
        with app.short_session(commit=True) as session:

            query = session.query(model.WebhookChange).order_by(
                model.WebhookChange.received
            )

            # nb. fetch (at most) 2 changes instead of just 1;
            # this will control time delay behavior below
            if changes := query[:2]:

                # process first change
                change = changes[0]
                log.info("processing webhook change: %s", change)
                processor.process_change(change)
                session.delete(change)

                # minimal time delay if 2nd change exists
                if len(changes) == 2:
                    time.sleep(0.1)
                    continue

        # nothing in queue, so wait 1 sec before checking again
        time.sleep(1)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS Medical Logs
"""

from wuttafarm.web.views.farmos.logs import LogMasterView
from wuttafarm.web.grids import SimpleSorter, StringFilter


class MedicalLogView(LogMasterView):
    """
    View for farmOS medical logs
    """

    model_name = "farmos_medical_log"
    model_title = "farmOS Medical Log"
    model_title_plural = "farmOS Medical Logs"

    route_prefix = "farmos_logs_medical"
    url_prefix = "/farmOS/logs/medical"

    farmos_log_type = "medical"
    farmos_refurl_path = "/logs/medical"

    labels = {
        "vet": "Veterinarian",
    }

    grid_columns = [
        "status",
        "drupal_id",
        "timestamp",
        "name",
        "assets",
        "vet",
        "owners",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # vet
        g.set_sorter("vet", SimpleSorter("vet"))
        g.set_filter("vet", StringFilter)

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # vet
        f.fields.insert_after("timestamp", "vet")


def defaults(config, **kwargs):
    base = globals()

    MedicalLogView = kwargs.get("MedicalLogView", base["MedicalLogView"])
    MedicalLogView.defaults(config)


def includeme(config):
    defaults(config)



"""add plant seasons

Revision ID: c5183b781d34
Revises: 5f474125a80e
Create Date: 2026-03-06 20:18:40.160531

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "c5183b781d34"
down_revision: Union[str, None] = "5f474125a80e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # season
    op.create_table(
        "season",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column("description", sa.String(length=255), nullable=True),
        sa.Column("farmos_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("drupal_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_season")),
        sa.UniqueConstraint("drupal_id", name=op.f("uq_season_drupal_id")),
        sa.UniqueConstraint("farmos_uuid", name=op.f("uq_season_farmos_uuid")),
        sa.UniqueConstraint("name", name=op.f("uq_season_name")),
    )
    op.create_table(
        "season_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column("name", sa.String(length=100), autoincrement=False, nullable=True),
        sa.Column(
            "description", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "farmos_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("drupal_id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_season_version")
        ),
    )
    op.create_index(
        op.f("ix_season_version_end_transaction_id"),
        "season_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_season_version_operation_type"),
        "season_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_season_version_pk_transaction_id",
        "season_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_season_version_pk_validity",
        "season_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_season_version_transaction_id"),
        "season_version",
        ["transaction_id"],
        unique=False,
    )

    # asset_plant_season
    op.create_table(
        "asset_plant_season",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("plant_asset_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("season_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["plant_asset_uuid"],
            ["asset_plant.uuid"],
            name=op.f("fk_asset_plant_season_plant_asset_uuid_asset_plant"),
        ),
        sa.ForeignKeyConstraint(
            ["season_uuid"],
            ["season.uuid"],
            name=op.f("fk_asset_plant_season_season_uuid_season"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_plant_season")),
    )
    op.create_table(
        "asset_plant_season_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "plant_asset_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "season_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_plant_season_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_plant_season_version_end_transaction_id"),
        "asset_plant_season_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_plant_season_version_operation_type"),
        "asset_plant_season_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_plant_season_version_pk_transaction_id",
        "asset_plant_season_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_plant_season_version_pk_validity",
        "asset_plant_season_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_plant_season_version_transaction_id"),
        "asset_plant_season_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # asset_plant_season
    op.drop_index(
        op.f("ix_asset_plant_season_version_transaction_id"),
        table_name="asset_plant_season_version",
    )
    op.drop_index(
        "ix_asset_plant_season_version_pk_validity",
        table_name="asset_plant_season_version",
    )
    op.drop_index(
        "ix_asset_plant_season_version_pk_transaction_id",
        table_name="asset_plant_season_version",
    )
    op.drop_index(
        op.f("ix_asset_plant_season_version_operation_type"),
        table_name="asset_plant_season_version",
    )
    op.drop_index(
        op.f("ix_asset_plant_season_version_end_transaction_id"),
        table_name="asset_plant_season_version",
    )
    op.drop_table("asset_plant_season_version")
    op.drop_table("asset_plant_season")

    # season
    op.drop_index(op.f("ix_season_version_transaction_id"), table_name="season_version")
    op.drop_index("ix_season_version_pk_validity", table_name="season_version")
    op.drop_index("ix_season_version_pk_transaction_id", table_name="season_version")
    op.drop_index(op.f("ix_season_version_operation_type"), table_name="season_version")
    op.drop_index(
        op.f("ix_season_version_end_transaction_id"), table_name="season_version"
    )
    op.drop_table("season_version")
    op.drop_table("season")



"""add WebhookChange

Revision ID: dd4d4142b96d
Revises: dca5b48a5562
Create Date: 2026-03-10 22:31:54.324952

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "dd4d4142b96d"
down_revision: Union[str, None] = "dca5b48a5562"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # webhook_change
    op.create_table(
        "webhook_change",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("entity_type", sa.String(length=100), nullable=False),
        sa.Column("bundle", sa.String(length=100), nullable=False),
        sa.Column("farmos_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("deleted", sa.Boolean(), nullable=False),
        sa.Column("received", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_webhook_change")),
    )


def downgrade() -> None:

    # webhook_change
    op.drop_table("webhook_change")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model definition for webhook changes
"""

import sqlalchemy as sa

from wuttjamaican.db import model
from wuttjamaican.util import make_utc


class WebhookChange(model.Base):
    """
    Represents a "change" (create/update/delete) notification which
    originated in farmOS and delivered via webhook.

    This table serves as a "FIFO queue" for processing the changes.
    """

    __tablename__ = "webhook_change"

    uuid = model.uuid_column()

    entity_type = sa.Column(sa.String(length=100), nullable=False)
    bundle = sa.Column(sa.String(length=100), nullable=False)
    farmos_uuid = sa.Column(model.UUID(), nullable=False)

    deleted = sa.Column(sa.Boolean(), nullable=False)

    received = sa.Column(
        sa.DateTime(),
        nullable=False,
        default=make_utc,
        doc="""
    Date and time when the change was obtained from the watcher thread.
    """,
    )

    def __str__(self):
        event_type = "delete" if self.deleted else "create/update"
        return f"{event_type} {self.entity_type}--{self.bundle}: {self.farmos_uuid}"



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm enum values
"""

from collections import OrderedDict

from wuttjamaican.enum import *


FARMOS_INTEGRATION_MODE_WRAPPER = "wrapper"
FARMOS_INTEGRATION_MODE_MIRROR = "mirror"
FARMOS_INTEGRATION_MODE_NONE = "none"

FARMOS_INTEGRATION_MODE = OrderedDict(
    [
        (FARMOS_INTEGRATION_MODE_WRAPPER, "wrapper (API only)"),
        (FARMOS_INTEGRATION_MODE_MIRROR, "mirror (2-way sync)"),
        (FARMOS_INTEGRATION_MODE_NONE, "none (standalone)"),
    ]
)


ANIMAL_SEX = OrderedDict(
    [
        ("M", "Male"),
        ("F", "Female"),
    ]
)


LOG_STATUS = OrderedDict(
    [
        ("pending", "Pending"),
        ("done", "Done"),
        ("abandoned", "Abandoned"),
    ]
)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for use as webhooks
"""

import logging
from uuid import UUID

from wuttaweb.views import View
from wuttaweb.db import Session


log = logging.getLogger(__name__)


class WebhookView(View):
    """
    Webhook views
    """

    def farmos_webhook(self):
        model = self.app.model
        session = Session()

        try:
            data = self.request.json
            log.debug("got webhook payload: %s", data)

            _, entity_type, event_type = data["event"].split(":")

            uuid = data["entity"]["uuid"][0]["value"]
            if entity_type == "taxonomy_term":
                bundle = data["entity"]["vid"][0]["target_id"]
            else:
                bundle = data["entity"]["type"][0]["target_id"]

            change = model.WebhookChange(
                entity_type=entity_type,
                bundle=bundle,
                farmos_uuid=UUID(uuid),
                deleted=event_type == "delete",
            )
            session.add(change)

        except:
            log.exception("failed to process webhook request")

        return {}

    @classmethod
    def defaults(cls, config):
        cls._webhook_defaults(config)

    @classmethod
    def _webhook_defaults(cls, config):

        # farmos webhook
        config.add_route("webhooks.farmos", "/farmos/webhook", request_method="POST")
        config.add_view(
            cls,
            attr="farmos_webhook",
            route_name="webhooks.farmos",
            require_csrf=False,
            renderer="json",
        )


def defaults(config, **kwargs):
    base = globals()

    WebhookView = kwargs.get("WebhookView", base["WebhookView"])
    WebhookView.defaults(config)


def includeme(config):
    defaults(config)
